/* mgd77_error_checking.h
 *
 * Contains definitions and functions useful for manipulation of mgd77 ascii
 * format data.
 *
 * Included are the following:
 *	int	mgd77_data_format_bad(string, killplus, ignore3)
 *	int	not_empty_nines(string, n)
 *	int	mgd77_data_read_rec(string, rec)
 *	int	data_has_valid_range(record)
 *	int	records_are_mergable(r1, r2, r3)
 *	int	mgd77_formatted_write(record, string)
 *	double	get_distance(lat1, lat2, lon1, lon2)
 *
 * Author:	Walter H. F. Smith
 * Date:	4 August, 1994
 * Version:	1.0
 *
 *	Modified: 14th march 1996 by Andrew Goodwillie:
 *
 *		"isdigit" test in function mgd77_data_format_bad()
 *		changed to "isprint" test to allow for spaces and
 *		letters in the seismic data fields.
 *
 */

/* 11-Jan-2005  kds  Switched nint to rint. */

/* On my sun, the man pages for strlen() and strncpy() tell me to do this:  */

#include <string.h>

/* On my sun, the man pages for isdigit() tell me to do this:  */

#include <ctype.h>

/* On my sun, the man pages for timegm() tell me to do this:  */

#include <time.h>

/* On my sun, the man pages for floor() tell me to do this, but you can also
define your own macro easily; GMT has one for floor()  */

#include <math.h>

/* On my sun, the man pages for sscanf() tell me to include this:  */

#include <stdio.h>

#include <stdlib.h> /* atoi() */

typedef int BOOLEAN;
#define TRUE 1
#define FALSE 0
#define D2R (M_PI / 180.0)
#define MAX_VELOCITY 50		/* Maximum expected velocity for a ship, in
					meters/second.  One knot is about
					0.515 m/s.  */

#define JUMP_DEGREES 1.0	/* For cruises which do not report time, give
					a warning when the change in lat or
					lon is >= to JUMP_DEGREES  */


#define MAX_BATHT 17.0		/* Max depth in two-way time seconds  */
#define MAX_BATHC 12750.0	/* Max depth in meters  */
#define MAX_MAGT 70000.0	/* Max mag total field intensity  */
#define MIN_MAGT 20000.0	/* Min mag total field intensity  */
#define MAX_MAGRES 8000.0	/* Max mag residual field (anomaly) */
#define MAX_MAGDIURN 500.0	/* Max mag diurnal correction  */
#define MAX_GRAVOBS 983650.0	/* Max absolute gravity  */
#define MIN_GRAVOBS 977600.0	/* Min absolute gravity  */
#define MAX_GRAVEOT 151.68	/* Correction value at 20 knots speed  */
#define MAX_GRAVFAA 400.0	/* Largest acceptable free-air anomaly */



struct MGD77_DATA_RECORD {
	/* The order of these elements is designed to minimize sizeof().
		The <name>_empty chars are set to 0 if valid data given.
		+1 if validly empty, and -1 if invalid data given.  One
		byte numbers in the record are stored as integers in a
		char, not as chars.  Thus the navigation quality code 
		indicating suspected by the data center is qual_n == 6,
		not qual_n == '6'  */
	double	lat;
	double	lon;
	double	batht;
	double	bathc;
	double	magt1;
	double	magt2;
	double	magres;
	double	magdiurn;
	double	magdepth;
	double	gravobs;
	double	graveot;
	double	gravfaa;
	char	cruiseid[9];
	char	time_empty;
	char	lat_empty;
	char	lon_empty;
	time_t	time;	/* Unix seconds since 1970.0 stored in GMT  */
	char	shotid[9];
	char	batht_empty;
	char	bathc_empty;
	char	bath_type_code;
	time_t	time_offset;	/* Amount that was added to time due to timezone;
			useful for checking if punched timezone should be ignored.  */
	short int	bath_corr_code;
	char	magt1_empty;
	char	magt2_empty;
	char	mag_res_sensor;
	char	magres_empty;
	char	magdiurn_empty;
	char	magdepth_empty;
	char	gravobs_empty;
	char	graveot_empty;
	char	gravfaa_empty;
	char	pos_type_code;
	char	qual_g;
	char	qual_m;
	char	qual_b;
	char	qual_n;
};

int	mgd77_data_format_bad();
int	not_empty_nines();
int	mgd77_data_read_rec();
int	data_has_valid_range();
int	mgd77_formatted_write();
double	get_distance();

	

int	mgd77_data_format_bad(string, killplus, ignore3)
char	string[];
BOOLEAN	killplus, ignore3;
{
	/* Returns 0 if the string appears to be in mgd77 data record format,
	and -1 if not.
	Requires that string[] be null-terminated like the usual C string.
	Thus strlen(string) should work, for example.

	If (killplus == TRUE) it will replace the '+' character in a sign
	column with the ' ' (space) character, which facilitates some char
	to numerical conversions.  Otherwise it will leave the string alone.

	If (ignore3 == TRUE) then it will not check if there is a '3' punched
	in the first char of string.  Otherwise it will check for this.   */

	int	i;
	BOOLEAN	error;

	if (strlen(string) != 120) return(-1);

	if (!ignore3 && string[0] != '3') return(-1);

	/* Check that digits and signs occur only where they should */
	i = 9;
	error = FALSE;
	while (i < 108 && !error) {
		switch (i) {
			case 9:
			case 27:
			case 35:
			case 72:
			case 79:
			case 84:
			case 97:
			case 103:
				error = !(string[i] == '+' || string[i] == '-' || string[i] == ' ');
				if (killplus && string[i] == '+') string[i] = ' ';
				break;
			default:
				error = !isdigit(string[i]);
				break;
		}
		i++;
	}
	if (error) return(-1);

	i = 108;
	while (i < 120 && !error) {
		error = !isprint(string[i]);
		i++;
	}

	if (error) return (-1);

	return(0);

}

int	not_empty_nines(string, n)
char	string[];
int	n;
{
	/* Given a character string, tests up to n chareturns beginning at
	string[0].  If at least one character in the string is not equal to
	'9' it returns 1, else it returns 0.  */

	int	i;

	i = 0;
	while (i < n) {
		if (string[i] != '9') return(1);
		i++;
	}
	return(0);
}

int	mgd77_data_read_rec(string, rec)
char	string[];
struct MGD77_DATA_RECORD *rec;
{
	/* Converts values in a 120-byte string to storage in structure.
	To be sure this will work, you should first call mgd77_data_format_bad()
	on the string.  This will return -1 if any data type decodes improperly;
	else it returns 0.  */

	char	tempstring[16];
	int	yr, mo, dy, hr, mn, sc, idummy, returned_value;
	time_t	tcoded;
	struct tm	tm;

	returned_value = 0;

	/* Get the Cruise ID and the Shot ID  */
	strncpy(rec->cruiseid, &string[1], 8);
	rec->cruiseid[8] = '\0';
	strncpy(rec->shotid, &string[108], 8);
	rec->shotid[8] = '\0';

	/* Get the time, if present:  */
	rec->time_offset = 0;
	if (not_empty_nines(&string[14], 13) ) {
		/* Time is supposedly coded in here.  */
		rec->time_empty = 0;
		rec->time_offset = 0;
		strncpy(tempstring, &string[14], 2);
		tempstring[2] = '\0';
		if ( (sscanf(tempstring, "%d", &yr)) != 1) {
			rec->time_empty = -1;
		}
		strncpy(tempstring, &string[16], 2);
		tempstring[2] = '\0';
		if ( (sscanf(tempstring, "%d", &mo)) != 1) {
			rec->time_empty = -1;
		}
		strncpy(tempstring, &string[18], 2);
		tempstring[2] = '\0';
		if ( (sscanf(tempstring, "%d", &dy)) != 1) {
			rec->time_empty = -1;
		}
		strncpy(tempstring, &string[20], 2);
		tempstring[2] = '\0';
		if ( (sscanf(tempstring, "%d", &hr)) != 1) {
			rec->time_empty = -1;
		}
		strncpy(tempstring, &string[22], 5);
		tempstring[5] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->time_empty = -1;
		}
		mn = idummy/1000;
		sc = (int)floor(0.06 * (idummy%1000));

		if (mo < 1 || mo > 12) rec->time_empty = -1;
		if (dy < 1 || dy > 31) rec->time_empty = -1;
		if (hr < 0 || hr > 23) rec->time_empty = -1;
		if (mn < 0 || mn > 59) rec->time_empty = -1;
		if (sc < 0 || sc > 59) rec->time_empty = -1;

		if (rec->time_empty == 0) {
			tm.tm_sec = sc;
			tm.tm_min = mn;
			tm.tm_hour = hr;
			tm.tm_mday = dy;
			tm.tm_mon = mo - 1;	/* struct tm uses 0-11 range  */
			tm.tm_year = yr;
			tm.tm_wday = tm.tm_yday = tm.tm_isdst = tm.tm_gmtoff = 0;
			tm.tm_zone = NULL;
			tcoded = timegm(&tm);

			/* Now try to get the time zone, if any:  */
			if (not_empty_nines(&string[10], 4) ) {
				strncpy(tempstring, &string[10], 4);
				tempstring[4] = '\0';
				if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
					rec->time_empty == -1;
				}
				else if (idummy != 0) {
					if (string[9] == '-') {
						rec->time_offset = -36 * idummy;
					}
					else {
						rec->time_offset = 36 * idummy;
					}
				}
			}
		}
	}
	else {
		rec->time_empty = 1;
	}
	rec->time = (rec->time_empty) ? 0 : tcoded + rec->time_offset;
	returned_value = (rec->time_empty == -1);


	/* Read lat and long, check ranges, read position type code.  */

	if (not_empty_nines(&string[28], 7) ) {
		strncpy(tempstring, &string[28], 7);
		tempstring[7] = 0;
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->lat_empty == -1;
			rec->lat = 0.0;
		}
		else {
			if (string[27] =='-') {
				rec->lat = -0.00001 * idummy;
			}
			else {
				rec->lat = 0.00001 * idummy;
			}
			rec->lat_empty = (idummy <= 9000000) ? 0 : -1;
		}
	}
	else {
		rec->lat_empty = 1;
		rec->lat = 0.0;
	}
	if (rec->lat_empty == -1) returned_value = -1;

	if (not_empty_nines(&string[36], 8) ) {
		strncpy(tempstring, &string[36], 8);
		tempstring[8] = 0;
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->lon_empty == -1;
			rec->lon = 0.0;
		}
		else {
			if (string[35] =='-') {
				if (idummy > 18000000) {
					rec->lon_empty = -1;
					rec->lon = 0.0;
				} 
				else {
					rec->lon = -0.00001 * idummy;
					rec->lon_empty = 0;
				}
			}
			else {
				if (idummy >= 36000000) {
					rec->lon_empty = -1;
					rec->lon = 0.0;
				}
				else {
					rec->lon = 0.00001 * idummy;
					rec->lon_empty = 0;
				}
			}
		}
	}
	else {
		rec->lon_empty = 1;
		rec->lon = 0.0;
	}
	if (rec->lon_empty == -1) returned_value = -1;

	rec->pos_type_code = string[44] - '0';

	/* Read bathymetry stuff:  */

	if (not_empty_nines(&string[45], 6) ) {
		strncpy(tempstring, &string[45], 6);
		tempstring[6] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->batht_empty == -1;
			rec->batht = 0.0;
			returned_value = -1;
		}
		else {
			rec->batht_empty = 0;
			rec->batht = 0.0001 * idummy;
		}
	}
	else {
		rec->batht_empty = 1;
		rec->batht = 0.0;
	}

	if (not_empty_nines(&string[51], 6) ) {
		strncpy(tempstring, &string[51], 6);
		tempstring[6] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->bathc_empty == -1;
			rec->bathc = 0.0;
			returned_value = -1;
		}
		else {
			rec->bathc_empty = 0;
			rec->bathc = 0.1 * idummy;
		}
	}
	else {
		rec->bathc_empty = 1;
		rec->bathc = 0.0;
	}

	strncpy(tempstring, &string[57], 2);
	tempstring[2] = '\0';
	rec->bath_corr_code = (short)atoi(tempstring);

	rec->bath_type_code = string[59] - '0';

	/* Read magnetic stuff:  */

	if (not_empty_nines(&string[60], 6) ) {
		strncpy(tempstring, &string[60], 6);
		tempstring[6] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->magt1_empty == -1;
			rec->magt1 = 0.0;
			returned_value = -1;
		}
		else {
			rec->magt1_empty = 0;
			rec->magt1 = 0.1 * idummy;
		}
	}
	else {
		rec->magt1_empty = 1;
		rec->magt1 = 0.0;
	}

	if (not_empty_nines(&string[66], 6) ) {
		strncpy(tempstring, &string[66], 6);
		tempstring[6] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->magt2_empty == -1;
			rec->magt2 = 0.0;
			returned_value = -1;
		}
		else {
			rec->magt2_empty = 0;
			rec->magt2 = 0.1 * idummy;
		}
	}
	else {
		rec->magt2_empty = 1;
		rec->magt2 = 0.0;
	}

	if (not_empty_nines(&string[73], 5) ) {
		strncpy(tempstring, &string[73], 5);
		tempstring[5] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->magres_empty == -1;
			rec->magres = 0.0;
			returned_value = -1;
		}
		else {
			rec->magres_empty = 0;
			rec->magres = (string[72] == '-') ? -0.1 * idummy : 0.1 * idummy;
		}
	}
	else {
		rec->magres_empty = 1;
		rec->magres = 0.0;
	}

	rec->mag_res_sensor = string[78] - '0';

	if (not_empty_nines(&string[80], 4) ) {
		strncpy(tempstring, &string[80], 4);
		tempstring[4] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->magdiurn_empty == -1;
			rec->magdiurn = 0.0;
			returned_value = -1;
		}
		else {
			rec->magdiurn_empty = 0;
			rec->magdiurn = (string[79] == '-') ? -0.1 * idummy : 0.1 * idummy;
		}
	}
	else {
		rec->magdiurn_empty = 1;
		rec->magdiurn = 0.0;
	}

	if (not_empty_nines(&string[85], 5) ) {
		strncpy(tempstring, &string[85], 5);
		tempstring[5] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->magdepth_empty == -1;
			rec->magdepth = 0.0;
			returned_value = -1;
		}
		else {
			rec->magdepth_empty = 0;
			rec->magdepth = (string[84] == '-') ? (double)(-idummy) : (double)idummy;
		}
	}
	else {
		rec->magdepth_empty = 1;
		rec->magdepth = 0.0;
	}

	/* Read gravity stuff:  */

	if (not_empty_nines(&string[90], 7) ) {
		strncpy(tempstring, &string[90], 7);
		tempstring[7] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->gravobs_empty == -1;
			rec->gravobs = 0.0;
			returned_value = -1;
		}
		else {
			rec->gravobs_empty = 0;
			rec->gravobs = 0.1 * idummy;
		}
	}
	else {
		rec->gravobs_empty = 1;
		rec->gravobs = 0.0;
	}

	if (not_empty_nines(&string[98], 5) ) {
		strncpy(tempstring, &string[98], 5);
		tempstring[5] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->graveot_empty == -1;
			rec->graveot = 0.0;
			returned_value = -1;
		}
		else {
			rec->graveot_empty = 0;
			rec->graveot = (string[97] == '-') ? -0.1 * idummy : 0.1 * idummy;
		}
	}
	else {
		rec->graveot_empty = 1;
		rec->graveot = 0.0;
	}

	if (not_empty_nines(&string[104], 4) ) {
		strncpy(tempstring, &string[104], 4);
		tempstring[4] = '\0';
		if ( (sscanf(tempstring, "%d", &idummy)) != 1) {
			rec->gravfaa_empty == -1;
			rec->gravfaa = 0.0;
			returned_value = -1;
		}
		else {
			rec->gravfaa_empty = 0;
			rec->gravfaa = (string[103] == '-') ? -0.1 * idummy : 0.1 * idummy;
		}
	}
	else {
		rec->gravfaa_empty = 1;
		rec->gravfaa = 0.0;
	}

	/* Read quality codes:  */

	rec->qual_g = string[116] - '0';
	rec->qual_m = string[117] - '0';
	rec->qual_b = string[118] - '0';
	rec->qual_n = string[119] - '0';

	return(returned_value);
}

int	data_has_valid_range(record, errstring)
struct MGD77_DATA_RECORD record;
char	errstring[];
{
	/* Returns 0 if the data fields (all but lat,lon,time) look OK.
	Returns -1 if any data fields are found to be bad which were not
	already flagged as such by the quality flags.  Returns 1 if one
	or more data fields are bad but all bad ones are flagged already.
	Writes a list of bad data field names into errstring, using lower
	case letters for fields already flagged, and upper case for fields
	not already flagged. errstring[] should be at least 84 bytes, or
	longer if it is going to have a message stored already when this
	is called.  It must be null terminated when this is called.

	*/

	int	nflagged = 0, nunflagged = 0;

	if (record.batht_empty == 0 && record.batht > MAX_BATHT) {
		if (record.qual_b == 7 || record.qual_b == 8) {
			nflagged++;
			strcat(errstring, "bath_t, ");
		}
		else {
			nunflagged++;
			strcat(errstring, "BATH_T, ");
		}
	}
	if (record.bathc_empty == 0 && record.bathc > MAX_BATHC) {
		if (record.qual_b == 7 || record.qual_b == 8) {
			nflagged++;
			strcat(errstring, "bath_c, ");
		}
		else {
			nunflagged++;
			strcat(errstring, "BATH_C, ");
		}
	}
	if (record.magt1_empty == 0 && (record.magt1 > MAX_MAGT || record.magt1 < MIN_MAGT) ) {
		if (record.qual_m == 0 || record.qual_m == 9) {
			nunflagged++;
			strcat(errstring, "MAG_T_1, ");
		}
		else {
			nflagged++;
			strcat(errstring, "mag_t_1, ");
		}
	}
	if (record.magt2_empty == 0 && (record.magt2 > MAX_MAGT || record.magt2 < MIN_MAGT) ) {
		if (record.qual_m == 0 || record.qual_m == 9) {
			nunflagged++;
			strcat(errstring, "MAG_T_2, ");
		}
		else {
			nflagged++;
			strcat(errstring, "mag_t_2, ");
		}
	}
	if (record.magres_empty == 0 && (fabs(record.magres) > MAX_MAGRES) ) {
		if (record.qual_m == 0 || record.qual_m == 9) {
			nunflagged++;
			strcat(errstring, "MAG_RES, ");
		}
		else {
			nflagged++;
			strcat(errstring, "mag_res, ");
		}
	}
	if (record.magdiurn_empty == 0 && (fabs(record.magdiurn) > MAX_MAGDIURN) ) {
		if (record.qual_m == 0 || record.qual_m == 9) {
			nunflagged++;
			strcat(errstring, "MAG_DIURN, ");
		}
		else {
			nflagged++;
			strcat(errstring, "mag_diurn, ");
		}
	}
	if (record.gravobs_empty == 0 && (record.gravobs > MAX_GRAVOBS || record.gravobs < MIN_GRAVOBS) ) {
		if (record.qual_g == 7 || record.qual_g == 8) {
			nflagged++;
			strcat(errstring, "grav_obs, ");
		}
		else {
			nunflagged++;
			strcat(errstring, "GRAV_OBS, ");
		}
	}
	if (record.graveot_empty == 0 && (fabs(record.graveot) > MAX_GRAVEOT) ) {
		if (record.qual_g == 7 || record.qual_g == 8) {
			nflagged++;
			strcat(errstring, "grav_eot, ");
		}
		else {
			nunflagged++;
			strcat(errstring, "GRAV_EOT, ");
		}
	}
	if (record.gravfaa_empty == 0 && (fabs(record.gravfaa) > MAX_GRAVFAA) ) {
		if (record.qual_g == 7 || record.qual_g == 8) {
			nflagged++;
			strcat(errstring, "grav_faa, ");
		}
		else {
			nunflagged++;
			strcat(errstring, "GRAV_FAA, ");
		}
	}

	/* Remove the extra comma and space at the end of the string:  */
	if (nunflagged + nflagged) errstring[strlen(errstring)-2] = '\0';

	if (nunflagged) return(-1);
	if (nflagged) return(1);
	return(0);
}


int	data_range_reset(record)
struct MGD77_DATA_RECORD *record;
{
	/* Returns 0 if the data fields (all but lat,lon,time) look OK.
	Returns -1 if any data fields are found to be out of range.  Any
	such points have their empty flag set to -1.  (Only those points
	with empty flag == 0 are checked.)  */

	int	range_errors = 0;

	if (record->batht_empty == 0 && record->batht > MAX_BATHT) {
		range_errors++;
		record->batht_empty = -1;
	}
	if (record->bathc_empty == 0 && record->bathc > MAX_BATHC) {
		range_errors++;
		record->bathc_empty = -1;
	}
	if (record->magt1_empty == 0 && (record->magt1 > MAX_MAGT || record->magt1 < MIN_MAGT) ) {
		range_errors++;
		record->magt1_empty = -1;
	}
	if (record->magt2_empty == 0 && (record->magt2 > MAX_MAGT || record->magt2 < MIN_MAGT) ) {
		range_errors++;
		record->magt2_empty = -1;
	}
	if (record->magres_empty == 0 && (fabs(record->magres) > MAX_MAGRES) ) {
		range_errors++;
		record->magres_empty = -1;
	}
	if (record->magdiurn_empty == 0 && (fabs(record->magdiurn) > MAX_MAGDIURN) ) {
		range_errors++;
		record->magdiurn_empty = -1;
	}
	if (record->gravobs_empty == 0 && (record->gravobs > MAX_GRAVOBS || record->gravobs < MIN_GRAVOBS) ) {
		range_errors++;
		record->gravobs_empty = -1;
	}
	if (record->graveot_empty == 0 && (fabs(record->graveot) > MAX_GRAVEOT) ) {
		range_errors++;
		record->graveot_empty = -1;
	}
	if (record->gravfaa_empty == 0 && (fabs(record->gravfaa) > MAX_GRAVFAA) ) {
		range_errors++;
		record->gravfaa_empty = -1;
	}

	if (range_errors) return(-1);
	return(0);
}

int	records_are_mergable(r1, r2, r3)
struct MGD77_DATA_RECORD r1, r2, *r3;
{
	/* Returns 1 if it is possible to merge r1 and r2, and 0 if not.
	During the course of this routine a partial merge is written in
	r3; if the merge fails and returns 0, there will be garbage in
	r3 (some new stuff and some old stuff).
	Generally, I expect this is only called if it has already been checked
	that r1 and r2 have the same time, lat, and lon.  The idea here is to
	patch together the data as needed. 
	This routine will take whichever record looks better.  It is possible
	to get an incompatible mixture, e.g. two-way-time from r1 and corrected
	depth from r2, but this only happens if the records are wierd.  */

	if (strncmp(r1.cruiseid, r2.cruiseid, 8)) 
		return(0);
	else {
		strncpy(r3->cruiseid, r1.cruiseid, 8);
		r3->cruiseid[8] = '\0';
	}

	if (strncmp(r1.shotid, r2.shotid, 8)) 
		return(0);
	else {
		strncpy(r3->shotid, r1.shotid, 8);
		r3->shotid[8] = '\0';
	}


	if (r1.time_empty == 0 && r2.time_empty == 0) {
		if (r1.time != r2.time) return(0);
		r3->time = r1.time;
		r3->time_offset = r1.time_offset;
		r3->time_empty = 0;
	} 
	else if (r1.time_empty == 0) {
		r3->time = r1.time;
		r3->time_offset = r1.time_offset;
		r3->time_empty = 0;
	}
	else if (r2.time_empty == 0) {
		r3->time = r2.time;
		r3->time_offset = r2.time_offset;
		r3->time_empty = 0;
	}
	else {
		r3->time = 0;
		r3->time_offset = 0;
		r3->time_empty = 1;
	}

	if (r1.lat_empty == 0 && r2.lat_empty == 0) {
		if (r1.lat != r2.lat) return(0);
		r3->lat = r1.lat;
		r3->lat_empty = 0;
	} 
	else if (r1.lat_empty == 0) {
		r3->lat = r1.lat;
		r3->lat_empty = 0;
	}
	else if (r2.lat_empty == 0) {
		r3->lat = r2.lat;
		r3->lat_empty = 0;
	}
	else {
		r3->lat = 0;
		r3->lat_empty = 1;
	}

	if (r1.lon_empty == 0 && r2.lon_empty == 0) {
		if (r1.lon != r2.lon) return(0);
		r3->lon = r1.lon;
		r3->lon_empty = 0;
	} 
	else if (r1.lon_empty == 0) {
		r3->lon = r1.lon;
		r3->lon_empty = 0;
	}
	else if (r2.lon_empty == 0) {
		r3->lon = r2.lon;
		r3->lon_empty = 0;
	}
	else {
		r3->lon = 0;
		r3->lon_empty = 1;
	}
	r3->qual_n = 9;
	if (r1.pos_type_code != r2.pos_type_code) return(0);
	r3->pos_type_code = r1.pos_type_code;

	if (r1.batht_empty == 0 && r2.batht_empty == 0) {
		if (r1.batht != r2.batht) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_b == r2.qual_b) return(0);
			if (r1.qual_b == 7 || r1.qual_b == 8) {
				r3->batht = r2.batht;
				r3->bath_type_code = r2.bath_type_code;
			}
			else {
				r3->batht = r1.batht;
				r3->bath_type_code = r1.bath_type_code;
			}
		}
		else {
			r3->batht = r1.batht;
			r3->bath_type_code = r1.bath_type_code;
		}
		r3->batht_empty = 0;
	} 
	else if (r1.batht_empty == 0) {
		r3->batht = r1.batht;
		r3->bath_type_code = r1.bath_type_code;
		r3->batht_empty = 0;
	}
	else if (r2.batht_empty == 0) {
		r3->batht = r2.batht;
		r3->bath_type_code = r2.bath_type_code;
		r3->batht_empty = 0;
	}
	else {
		r3->batht = 0;
		r3->batht_empty = 1;
		r3->bath_type_code = 9;
	}

	if (r1.bathc_empty == 0 && r2.bathc_empty == 0) {
		if (r1.bathc != r2.bathc) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_b == r2.qual_b) return(0);
			if (r1.qual_b == 7 || r1.qual_b == 8) {
				r3->bathc = r2.bathc;
				r3->bath_corr_code = r2.bath_corr_code;
			}
			else {
				r3->bathc = r1.bathc;
				r3->bath_corr_code = r1.bath_corr_code;
			}
		}
		else {
			r3->bathc = r1.bathc;
			r3->bath_corr_code = r1.bath_corr_code;
		}
		r3->bathc_empty = 0;
	} 
	else if (r1.bathc_empty == 0) {
		r3->bathc = r1.bathc;
		r3->bath_corr_code = r1.bath_corr_code;
		r3->bathc_empty = 0;
	}
	else if (r2.bathc_empty == 0) {
		r3->bathc = r2.bathc;
		r3->bath_corr_code = r2.bath_corr_code;
		r3->bathc_empty = 0;
	}
	else {
		r3->bathc = 0;
		r3->bath_corr_code = 99;
		r3->bathc_empty = 1;
	}
	r3->qual_b = 9;
	
	if (r1.magt1_empty == 0 && r2.magt1_empty == 0) {
		if (r1.magt1 != r2.magt1) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_m == r2.qual_m) return(0);
			if (r1.qual_m == 7 || r1.qual_m == 8)
				r3->magt1 = r2.magt1;
			else
				r3->magt1 = r1.magt1;
		}
		else
			r3->magt1 = r1.magt1;
		r3->magt1_empty = 0;
	} 
	else if (r1.magt1_empty == 0) {
		r3->magt1 = r1.magt1;
		r3->magt1_empty = 0;
	}
	else if (r2.magt1_empty == 0) {
		r3->magt1 = r2.magt1;
		r3->magt1_empty = 0;
	}
	else {
		r3->magt1 = 0;
		r3->magt1_empty = 1;
	}

	if (r1.magt2_empty == 0 && r2.magt2_empty == 0) {
		if (r1.magt2 != r2.magt2) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_m == r2.qual_m) return(0);
			if (r1.qual_m == 7 || r1.qual_m == 8)
				r3->magt2 = r2.magt2;
			else
				r3->magt2 = r1.magt2;
		}
		else
			r3->magt2 = r1.magt2;
		r3->magt2_empty = 0;
	} 
	else if (r1.magt2_empty == 0) {
		r3->magt2 = r1.magt2;
		r3->magt2_empty = 0;
	}
	else if (r2.magt2_empty == 0) {
		r3->magt2 = r2.magt2;
		r3->magt2_empty = 0;
	}
	else {
		r3->magt2 = 0;
		r3->magt2_empty = 1;
	}

	if (r1.magres_empty == 0 && r2.magres_empty == 0) {
		if (r1.magres != r2.magres) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_m == r2.qual_m) return(0);
			if (r1.qual_m == 7 || r1.qual_m == 8)
				r3->magres = r2.magres;
			else
				r3->magres = r1.magres;
		}
		else
			r3->magres = r1.magres;
		r3->magres_empty = 0;
	} 
	else if (r1.magres_empty == 0) {
		r3->magres = r1.magres;
		r3->magres_empty = 0;
	}
	else if (r2.magres_empty == 0) {
		r3->magres = r2.magres;
		r3->magres_empty = 0;
	}
	else {
		r3->magres = 0;
		r3->magres_empty = 1;
	}

	if (r1.magdiurn_empty == 0 && r2.magdiurn_empty == 0) {
		if (r1.magdiurn != r2.magdiurn) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_m == r2.qual_m) return(0);
			if (r1.qual_m == 7 || r1.qual_m == 8)
				r3->magdiurn = r2.magdiurn;
			else
				r3->magdiurn = r1.magdiurn;
		}
		else
			r3->magdiurn = r1.magdiurn;
		r3->magdiurn_empty = 0;
	} 
	else if (r1.magdiurn_empty == 0) {
		r3->magdiurn = r1.magdiurn;
		r3->magdiurn_empty = 0;
	}
	else if (r2.magdiurn_empty == 0) {
		r3->magdiurn = r2.magdiurn;
		r3->magdiurn_empty = 0;
	}
	else {
		r3->magdiurn = 0;
		r3->magdiurn_empty = 1;
	}

	if (r1.magdepth_empty == 0 && r2.magdepth_empty == 0) {
		if (r1.magdepth != r2.magdepth) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_m == r2.qual_m) return(0);
			if (r1.qual_m == 7 || r1.qual_m == 8)
				r3->magdepth = r2.magdepth;
			else
				r3->magdepth = r1.magdepth;
		}
		else
			r3->magdepth = r1.magdepth;
		r3->magdepth_empty = 0;
	} 
	else if (r1.magdepth_empty == 0) {
		r3->magdepth = r1.magdepth;
		r3->magdepth_empty = 0;
	}
	else if (r2.magdepth_empty == 0) {
		r3->magdepth = r2.magdepth;
		r3->magdepth_empty = 0;
	}
	else {
		r3->magdepth = 0;
		r3->magdepth_empty = 1;
	}
	r3->qual_m = 9;

	if (r1.gravobs_empty == 0 && r2.gravobs_empty == 0) {
		if (r1.gravobs != r2.gravobs) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_g == r2.qual_g) return(0);
			if (r1.qual_g == 7 || r1.qual_g == 8)
				r3->gravobs = r2.gravobs;
			else
				r3->gravobs = r1.gravobs;
		}
		else
			r3->gravobs = r1.gravobs;
		r3->gravobs_empty = 0;
	} 
	else if (r1.gravobs_empty == 0) {
		r3->gravobs = r1.gravobs;
		r3->gravobs_empty = 0;
	}
	else if (r2.gravobs_empty == 0) {
		r3->gravobs = r2.gravobs;
		r3->gravobs_empty = 0;
	}
	else {
		r3->gravobs = 0;
		r3->gravobs_empty = 1;
	}

	if (r1.graveot_empty == 0 && r2.graveot_empty == 0) {
		if (r1.graveot != r2.graveot) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_g == r2.qual_g) return(0);
			if (r1.qual_g == 7 || r1.qual_g == 8)
				r3->graveot = r2.graveot;
			else
				r3->graveot = r1.graveot;
		}
		else
			r3->graveot = r1.graveot;
		r3->graveot_empty = 0;
	} 
	else if (r1.graveot_empty == 0) {
		r3->graveot = r1.graveot;
		r3->graveot_empty = 0;
	}
	else if (r2.graveot_empty == 0) {
		r3->graveot = r2.graveot;
		r3->graveot_empty = 0;
	}
	else {
		r3->graveot = 0;
		r3->graveot_empty = 1;
	}

	if (r1.gravfaa_empty == 0 && r2.gravfaa_empty == 0) {
		if (r1.gravfaa != r2.gravfaa) {
			/* Does one have a better quality code  ?  */
			if (r1.qual_g == r2.qual_g) return(0);
			if (r1.qual_g == 7 || r1.qual_g == 8)
				r3->gravfaa = r2.gravfaa;
			else
				r3->gravfaa = r1.gravfaa;
		}
		else
			r3->gravfaa = r1.gravfaa;
		r3->gravfaa_empty = 0;
	} 
	else if (r1.gravfaa_empty == 0) {
		r3->gravfaa = r1.gravfaa;
		r3->gravfaa_empty = 0;
	}
	else if (r2.gravfaa_empty == 0) {
		r3->gravfaa = r2.gravfaa;
		r3->gravfaa_empty = 0;
	}
	else {
		r3->gravfaa = 0;
		r3->gravfaa_empty = 1;
	}
	r3->qual_g = 9;

	return(1);
}

int	mgd77_formatted_write(record, string)
struct MGD77_DATA_RECORD record;
char	string[];
{
	/* Formats the binary data in record into string.

	Right now, this always returns 0, but maybe later 
	I will build error checking into it.  */

	int	idummy, j;
	struct tm *tm;
	time_t	t;

	string[0] = '3';

	strncpy(&string[1], record.cruiseid, 8);

	if (record.time_empty == 0) {

		string[9] = (record.time_offset >= 0) ? '+' : '-';
		idummy = abs(record.time_offset)/36;
		sprintf(&string[10], "%04d", idummy);

		t = record.time - record.time_offset;
		tm = gmtime(&t);
		idummy = tm->tm_year;
		sprintf(&string[14], "%02d", idummy);
		idummy = tm->tm_mon + 1;
		sprintf(&string[16], "%02d", idummy);
		idummy = tm->tm_mday;
		sprintf(&string[18], "%02d", idummy);
		idummy = tm->tm_hour;
		sprintf(&string[20], "%02d", idummy);
		idummy = tm->tm_min*1000 + (int)floor(tm->tm_sec *1000.0 / 60.0);
		sprintf(&string[22], "%05d", idummy);
	}
	else {
		string[9] = '+';
		for (j = 10; j < 27; j++) string[j] = '9';
	}

	if (record.lat_empty == 0) {
		string[27] = (record.lat >= 0) ? '+' : '-';
		idummy = rint(fabs(record.lat) * 1.0e+05);
		sprintf(&string[28], "%07d", idummy);
	}
	else {
		string[27] = '+';
		for (j = 28; j < 35; j++) string[j] = '9';
	}

	if (record.lon_empty == 0) {
		string[35] = (record.lon >= 0) ? '+' : '-';
		idummy = rint(fabs(record.lon) * 1.0e+05);
		sprintf(&string[36], "%08d", idummy);
	}
	else {
		string[35] = '+';
		for (j = 37; j < 44; j++) string[j] = '9';
	}

	idummy = (int)record.pos_type_code;
	sprintf(&string[44], "%01d", idummy);

	if (record.batht_empty == 0) {
	  idummy = rint(record.batht * 1.0e+04);
		sprintf(&string[45], "%06d", idummy);
	}
	else {
		for (j = 45; j < 51; j++) string[j] = '9';
	}

	if (record.bathc_empty == 0) {
	  idummy = rint(record.bathc * 10.0);
		sprintf(&string[51], "%06d", idummy);
	}
	else {
		for (j = 51; j < 57; j++) string[j] = '9';
	}

	idummy = record.bath_corr_code;
	sprintf(&string[57], "%02d", idummy);

	idummy = record.bath_type_code;
	sprintf(&string[59], "%01d", idummy);

	if (record.magt1_empty == 0) {
	  idummy = rint(record.magt1 * 10.0);
	  sprintf(&string[60], "%06d", idummy);
	}
	else {
		for (j = 60; j < 66; j++) string[j] = '9';
	}

	if (record.magt2_empty == 0) {
	  idummy = rint(record.magt2 * 10.0);
	  sprintf(&string[66], "%06d", idummy);
	}
	else {
		for (j = 66; j < 72; j++) string[j] = '9';
	}
	
	if (record.magres_empty == 0) {
	  idummy = rint(fabs(record.magres) * 10.0);
		string[72] = (record.magres >= 0.0) ? '+' : '-';
		sprintf(&string[73], "%05d", idummy);
	}
	else {
		string[72] = '+';
		for (j = 73; j < 78; j++) string[j] = '9';
	}

	idummy = record.mag_res_sensor;
	sprintf(&string[78], "%01d", idummy);

	if (record.magdiurn_empty == 0) {
	  idummy = rint(fabs(record.magdiurn) * 10.0);
		string[79] = (record.magdiurn >= 0.0) ? '+' : '-';
		sprintf(&string[80], "%04d", idummy);
	}
	else {
		string[79] = '+';
		for (j = 80; j < 84; j++) string[j] = '9';
	}

	if (record.magdepth_empty == 0) {
	  idummy = rint(fabs(record.magdepth));
		string[84] = (record.magdepth >= 0.0) ? '+' : '-';
		sprintf(&string[80], "%05d", idummy);
	}
	else {
		string[84] = '+';
		for (j = 85; j < 90; j++) string[j] = '0';
	}

	if (record.gravobs_empty == 0) {
	  idummy = rint(record.gravobs * 10.0);
		sprintf(&string[90], "%07d", idummy);
	}
	else {
		for (j = 90; j < 97; j++) string[j] = '9';
	}

	if (record.graveot_empty == 0) {
		idummy = rint(fabs(record.graveot) * 10.0);
		string[97] = (record.graveot >= 0.0) ? '+' : '-';
		sprintf(&string[98], "%05d", idummy);
	}
	else {
		string[97] = '+';
		for (j = 98; j < 103; j++) string[j] = '9';
	}

	if (record.gravfaa_empty == 0) {
		idummy = rint(fabs(record.gravfaa) * 10.0);
		string[103] = (record.gravfaa >= 0.0) ? '+' : '-';
		sprintf(&string[104], "%04d", idummy);
	}
	else {
		string[103] = '+';
		for (j = 104; j < 108; j++) string[j] = '9';
	}

	strncpy(&string[108], record.shotid, 8);

	idummy = record.qual_g;
	sprintf(&string[116], "%01d", idummy);
	idummy = record.qual_m;
	sprintf(&string[117], "%01d", idummy);
	idummy = record.qual_b;
	sprintf(&string[118], "%01d", idummy);
	idummy = record.qual_n;
	sprintf(&string[119], "%01d", idummy);

	string[120] = '\0';

	return(0);
}

double	get_distance(lat1, lat2, lon1, lon2)
double	lat1, lat2, lon1, lon2;
{
	/* Return approximate distance in meters  */

	double	clon1, clon2, clat1, clat2;
	double	slon1, slon2, slat1, slat2;
	double	x1, y1, z1, x2, y2, z2;


	if (fabs(lat1-lat2) < 2.0) {
		/* Pretend the Earth is flat  */
		clat1 = cos(D2R * 0.5 * (lat1 + lat2));
		x1 = fabs(lon2-lon1);
		if (x1 > 180.0) x1 = 360.0 - x1;
		return(111194.9 * hypot((lat2-lat1), clat1*x1));
	}
	
	clon1 = cos(D2R * lon1);
	clon2 = cos(D2R * lon2);
	clat1 = cos(D2R * lat1);
	clat2 = cos(D2R * lat2);
	slon1 = sin(D2R * lon1);
	slon2 = sin(D2R * lon2);
	slat1 = sin(D2R * lat1);
	slat2 = sin(D2R * lat2);

	x1 = clon1*clat1;
	y1 = slon1*clat1;
	z1 = slat1;

	x2 = clon2*clat2;
	y2 = slon2*clat2;
	z2 = slat2;

	return(6371000.0*acos(x1*x2 + y1*y2 + z1*z2));
}



