/*
 * mgd77_good_bad reads a mgd77 format file such as out of the NGDC GEODAS CD-ROM
 * and looks for trouble spots.  The usage is
 *
 * mgd77_good_bad <list_in> -G<list_good> -B<list_bad> [-V]
 *
 * Input:  <list_in> = filename for a file containing filenames to check,
 * 	one per line, including the directory path if needed.
 *
 * Output:	<list_good> and <list_bad> have the filenames from <list_in>
 * 	and a report of any problems found.
 *
 *
 * Author:	Walter H. F. Smith
 * Date:	5 August, 1994
 *
 */

#include "mgd77_error_checking.h"

int
main (int argc, char *argv[])
{
	char line[512], name[512], errstring[128];
	BOOLEAN	error, quick_exit, verbose;
	int	i, nheadrecs, ndatarecs, line_no, reclength, seq_no;
	int	nduplicates, ntimegiven, nunflaggednav, nflaggednav, nflaggeddata, nunflaggeddata;
	FILE *fpinput, *fpgood, *fpbad, *fpmgd77;
	time_t	timenow;
	double	distance, speed;
	struct MGD77_DATA_RECORD rec, lastrec, mergerec;



	fpgood = fpbad = fpinput = NULL;
	error = verbose = quick_exit = FALSE;

	for (i = 1; !error && i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case '\0':
					quick_exit = TRUE;
					break;
				case 'B':
					if ( (fpbad = fopen(&argv[i][2], "w")) == NULL) {
						fprintf(stderr,"mgd77_good_bad:  ERROR cannot open w %s\n", &argv[i][2]);
						error = TRUE;
					}
					break;
				case 'G':
					if ( (fpgood = fopen(&argv[i][2], "w")) == NULL) {
						fprintf(stderr,"mgd77_good_bad:  ERROR cannot open w %s\n", &argv[i][2]);
						error = TRUE;
					}
					break;
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else {
			if ( (fpinput = fopen (argv[i], "r")) == NULL) {
				fprintf(stderr,"mgd77_good_bad:  ERROR cannot open r %s\n", &argv[i]);
				error = TRUE;
			}
		}
	}

	if (fpgood == NULL || fpbad == NULL) error = TRUE;
	
	if (argc == 1 || error) {
		fprintf (stderr, "usage: mgd77_good_bad <listfile> -B<badfile> -G<goodfile> [-V]\n\n");
		if (quick_exit) exit(-1);
		fprintf (stderr, "\t<listfile> is the filename for an input file listing mgd77 files to check.\n");
		fprintf (stderr, "\t\tIf listfile is not specified, mgd77_good_bad will look on stdin.\n");
		fprintf (stderr, "\t<badfile> is the filename for an output file listing the mgd77 files which have problems.\n");
		fprintf (stderr, "\t<goodfile> is the filename for an output file listing the mgd77 files which are OK.\n");
		fprintf (stderr, "mgd77_good_bad optional arguments:\n");
		fprintf (stderr, "\t-V will report progress.  [Default runs silently.]\n");
		exit (-1);
	}

	if (fpinput == NULL) fpinput = stdin;

	timenow = time((time_t *)0);

	while (fgets (line, 512, fpinput)) {
		line[strlen(line)-1] = '\0';
		if ( (fpmgd77 = fopen(line, "r") ) == NULL) {
			fprintf(stderr,"mgd77_good_bad:  ERROR cannot open r %s\n", line);
			fprintf(fpbad,"%s could not be opened for checking.\n", line);
			continue;
		}
		if (verbose) fprintf(stderr,"mgd77_good_bad thinks %s is", line);

		nheadrecs = ndatarecs = 0;
		nduplicates = nflaggednav = nunflaggednav = 0;
		ntimegiven = nflaggeddata = nunflaggeddata = 0;
		line_no = 0;
		strcpy(name, line);
		error = FALSE;

		while (!error && (fgets (line, 512, fpmgd77))) {
			line_no++;
			reclength = strlen(line);
			line[reclength-1] = '\0';
			reclength--;
			if (reclength == 80) {
				nheadrecs++;
				if (ndatarecs) {
					/* This is bad.  A header record has come after a data record.  */
					error = TRUE;
					fprintf(fpbad, "%s has a header record after a data record at line %ld.\n", name, line_no);
					if (verbose) fprintf(stderr, " bad.\n");
				}	
				else {
					/* Check if last two columns of header are formatted correctly:  */
					if ( (sscanf(&line[78], "%d", &seq_no)) != 1 || seq_no != nheadrecs) {
						error = TRUE;
						fprintf(fpbad, "%s has a header record sequence error at line %ld.\n", name, line_no);
						if (verbose) fprintf(stderr, " bad.\n");
					}
					else if (nheadrecs == 1 && line[0] != '1') {
						error = TRUE;
						fprintf(fpbad, "%s first header record not marked with a 1 at line %ld.\n", name, line_no);
						if (verbose) fprintf(stderr, " bad.\n");
					}
				}
			}
			else if (reclength == 120) {
				ndatarecs++;
				errstring[0] = '\0';
				if ( (mgd77_data_format_bad(line, 0, 0) ) || (mgd77_data_read_rec(line, &rec) ) ) {
					error = TRUE;
					fprintf(fpbad, "%s data record format unreadable at line %ld.\n", name, line_no);
					if (verbose) fprintf(stderr, " bad.\n");
					continue;
				}
				if (error = (rec.lat_empty == 1 || rec.lon_empty == 1) ) {
					fprintf(fpbad, "%s data record missing lat or lon at line %ld.\n", name, line_no);
					if (verbose) fprintf(stderr, " bad.\n");
					continue;
				}
				if (error = (rec.time_empty == 0 && rec.time > timenow) ) {
					fprintf(fpbad, "%s data record shows future time at line %ld.\n", name, line_no);
					if (verbose) fprintf(stderr, " bad.\n");
					continue;
				}
				if (ndatarecs > 1) {
					/* Could check gradients here.  */
					if (lastrec.time_empty == 0 && rec.time_empty == 0) {
						if (error = (rec.time < lastrec.time) ) {
							fprintf(fpbad, "%s time decreases at line %ld.\n", name, line_no);
							if (verbose) fprintf(stderr, " bad.\n");
							continue;
						}
						if (rec.time == lastrec.time) {
							if (records_are_mergable(lastrec, rec, &mergerec) ) {
								nduplicates++;
							}
							else {
								error = TRUE;
								fprintf(fpbad, "%s has unmergable duplicate time at line %ld.\n", name, line_no);
								if (verbose) fprintf(stderr, " bad.\n");
								continue;
							}
						}
						distance = get_distance(rec.lat, lastrec.lat, rec.lon, lastrec.lon);
						speed = distance / (rec.time - lastrec.time);
						if (speed > MAX_VELOCITY) {
							if (lastrec.qual_n != 9 || rec.qual_n != 9)
								nflaggednav++;
							else
								nunflaggednav++;
						}
					}
				}
				if (error = data_has_valid_range(rec, errstring) ) {
					if (error > 0)
						nflaggeddata++;
					else
						nunflaggeddata++;
					error = 0;
				}
				if (rec.time_empty == 0) ntimegiven++;
				lastrec = rec;
			}
			else {
				
				/*  A record with neither 80 nor 120 bytes is found  */
				error = TRUE;
				fprintf(fpbad, "%s has record of neither 80 nor 120 bytes at line %ld.\n", name, line_no);
				if (verbose) fprintf(stderr, " bad.\n");
			}
		}
		fclose(fpmgd77);

		/* When we get here, if error == TRUE we bombed on only a partial read, and already wrote fpbad and verbose.
			If error == FALSE, we need to check time status and other error status before pronouncing file clean.
		*/
		if (!error) {
			/* Store n time missing in ntimegiven:  */
			ntimegiven = ndatarecs - ntimegiven;
			if (nflaggednav || nunflaggednav || nflaggeddata || nunflaggeddata || nduplicates || ntimegiven) {
				if (verbose) fprintf(stderr, " bad.\n");
fprintf(fpbad, "%s nhrecs=%d ndrecs=%d ndupes=%d ntimemissing=%d nflaggednav=%d nunflaggednav=%d nflaggeddat=%d nunflaggeddat=%d.\n",
	name, nheadrecs, ndatarecs, nduplicates, ntimegiven, nflaggednav, nunflaggednav, nflaggeddata, nunflaggeddata);	
			}
			else {
				if (verbose) fprintf(stderr, " good.\n");	
				if (nheadrecs%24) {
					fprintf(fpgood, "%s looks clean.  nheadrecs = %d, ndatarecs = %d. WARNING:  nheadrecs mod 24 not zero.\n",
						name, nheadrecs, ndatarecs);
				}
				else {
					fprintf(fpgood, "%s looks clean.  nheadrecs = %d, ndatarecs = %d.\n", name, nheadrecs, ndatarecs);
				}
			}
		}
	}
	fclose(fpinput);
	exit(0);
}
